#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <GL/glut.h>
#include "graph.h"

#define MAX_DISPARITY 64
#define LARGE_VALUE 100
#define HUGE_VALUE 16777216

template class Graph<int,int,int>;
template class Graph<short,int,int>;
template class Graph<float,float,float>;
template class Graph<double,double,double>;

typedef Graph<float,float,float> GraphType;

typedef unsigned char uchar;

const int bits[3]={1,1,3};

typedef struct _Pict{
  int width;
  int height;
  int depth; 
  uchar *image;
  GLenum format;
  int type;
} Pict;

Pict* ReadImage_PNM(char* fname)
{
  FILE *fp;
  char str[256];
  Pict* pic=(Pict*)malloc(sizeof(Pict));
  int type;
  GLenum format[6]={GL_LUMINANCE,GL_LUMINANCE,GL_RGB,
		    GL_LUMINANCE,GL_LUMINANCE,GL_RGB};
  fp=fopen(fname,"r");
  if(fp==NULL){
    printf("File Open Error!!\n");
    exit(-1);
  }
  
  fgets(str,256,fp);
  
  // if(str[0]=='#')
  //fgets(str,256,fp);

  sscanf(str,"P%d",&type);
  pic->format=format[type-1];
  pic->type=type;

  if(type<1 && type>6){
    printf("Not Match File Type!!\n");
    exit(-1);
  }


  fgets(str,256,fp);

  if(str[0]=='#')
    fgets(str,256,fp);

  if(type==2||type==3||type==5||type==6){

    switch(sscanf(str,"%d %d %d",&pic->width,&pic->height,&pic->depth)){

    case 1:
      fgets(str,256,fp);
      if(sscanf(str,"%d %d",&pic->height,&pic->depth)==1){
	fgets(str,256,fp);
	sscanf(str,"%d",&pic->depth);
      }
      break;
    case 2:
      fgets(str,256,fp);
      sscanf(str,"%d",&pic->depth);
      break;
    }
  }else if(type==1||type==4){
    switch(sscanf(str,"%d %d",&pic->width,&pic->height)){

    case 1:
      fgets(str,256,fp);
      sscanf(str,"%d",&pic->height);
      break;

    }

  }


  pic->image=(uchar*)malloc(sizeof(uchar)*pic->width*pic->height*bits[(type-1)%3]);

  if(type<=3){
    for(int i=0;i<pic->height;i++){
      for(int j=0;j<pic->width;j++){
	for(int k=0;k<bits[(type-1)%3];k++){
	  fscanf(fp,"%d",
		 &pic->image[pic->width*bits[(type-1)%3]*i+
			     bits[(type-1)%3]*j+k]);
	}
      }
    }
  }else if(type>3){
    fread(pic->image,sizeof(uchar),pic->height*pic->width*bits[(type-1)%3],fp);
  }
  
  fclose(fp);

  return pic;

}


void WriteImage_PNM(Pict* pic,char* fname){
  FILE *fp;
  

  if((fp=fopen(fname,"w"))==NULL){
    printf("Directory Not Exist!!\n");
    exit(-1);
  }
  
  fprintf(fp,"P%d\n",pic->type);
  fprintf(fp,"# CREATOR: Tatsuya\n");
  fprintf(fp,"%d %d\n",pic->width,pic->height);
  
  if(pic->type==2 || pic->type==3 || pic->type==5 || pic->type==6)
    fprintf(fp,"%d\n",pic->depth);
  
  if(pic->type<=3){
    //    for(int i=0;i<pic->width*pic->height*bits[(type-1)%3];i++){
    for(int i=0;i<pic->height;i++){
      for(int j=0;j<pic->width;j++){
	for(int k=0;k<bits[(pic->type-1)%3];k++){
	  fprintf(fp,"%d ",pic->image[pic->width*
				      bits[(pic->type-1)%3]*i+
				      bits[(pic->type-1)%3]*j+k]);
		  
	  
	}

	fprintf(fp,"\n");
	//	if(j%4==0 && j!=0)
	// fprintf(fp,"\n");

      }	
      
    }
  }else if(pic->type>3){
    fwrite(pic->image,sizeof(uchar),
	   pic->height*pic->width*bits[(pic->type-1)%3],fp);
  }

   fclose(fp);

}

void moge(double n)
{
  fprintf(stderr,"moge(%f)\n",n);
}

void moge(int n,int m)
{
  fprintf(stderr,"moge(%d,%d)\n",n,m);
}

void moge(char *str,int n)
{
  fprintf(stderr,"moge(%s:%d)\n",str,n);
}


int Interval(void)
{
  static unsigned int pre;
  unsigned int c,interval;

  c=clock();
  interval=c-pre;
  pre=c;

  if(interval<10000)
    return 10000;
  else
    return interval;
}

double Max(double a,double b)
{
  if(a>=b)
    return a;
  else
    return b;
}

uchar *Alloc_1Dub(int w)
{
  uchar *data;

  data=(uchar*)malloc(sizeof(uchar)*w);

  return data;
}

int *Alloc_1Di(int w)
{
  int *data;

  data=(int*)malloc(sizeof(int)*w);

  return data;
}

double *Alloc_1Dd(int w)
{
  double *data;

  data=(double*)malloc(sizeof(double)*w);

  return data;
}

GLfloat *Alloc_1Dglf(int w)
{
  GLfloat *data;

  data=(GLfloat*)malloc(sizeof(GLfloat)*w);

  return data;
}

uchar **Alloc_2Dub(int w,int h)
{
  uchar *data;
  uchar **pdata;

  data=(uchar*)malloc(sizeof(uchar)*w*h);

  pdata=(uchar**)malloc(sizeof(uchar*)*h);
  for(int i=0;i<h;++i)
    pdata[i]=&data[i*w];

  for(int i=0;i<w*h;++i)
    data[i]=0;

  return pdata;
}

int **Alloc_2Di(int w,int h)
{
  int *data;
  int **pdata;

  data=(int*)malloc(sizeof(int)*w*h);

  pdata=(int**)malloc(sizeof(int*)*h);
  for(int i=0;i<h;++i)
    pdata[i]=&data[i*w];

  for(int i=0;i<w*h;++i)
    data[i]=0;

  return pdata;
}

double **Alloc_2Dd(int w,int h)
{
  double *data;
  double **pdata;

  data=(double*)malloc(sizeof(double)*w*h);

  pdata=(double**)malloc(sizeof(double*)*h);
  for(int i=0;i<h;++i)
    pdata[i]=&data[i*w];

  for(int i=0;i<w*h;++i)
    data[i]=0.0;

  return pdata;
}


double Get_Luminance(uchar r,uchar g,uchar b)
{
  return 0.2990*(double)r+0.5864*(double)g+0.1146*(double)b;
}

double Get_Data_Cost(double a)
{
  double weight;

  weight=0.5;

  return weight*a;
}

double Get_Smooth_Cost(double a,double b)
{
  double weight;

  if(fabs(a-b)<0.000000001)
    return 0.0;

  weight=3.0;

  return weight*fabs(a-b);
}

int Get_Alpha_Node(int kind,int x,int y,
			  int img_width,int img_height)
{
  int ret_value=0;

  switch(kind)
    {
    case 0:
      {
	ret_value=img_width*y+x;
	break;
      }
    case 1://****xåΡ****//
      {
	ret_value=(img_width*img_height)+
	  ((img_width-1)*y)+x;
	break;
      }
    case 2://****yåΡ****//
      {
	ret_value=(img_width*img_height)+
	  ((img_width-1)*img_height)+
	  ((img_height-1)*x)+y;
	break;
      }
    default:
      break;
    }

  return ret_value;
}

int **Calc_Disparity_Array(int n)
{
  int **ret_value;
  int count;

  ret_value=Alloc_2Di(n,n);

  for(int i=0;i<n;++i)
    for(int j=0;j<n;++j)  
      ret_value[i][j]=-1;

  ret_value[0][0]=0;
  ret_value[1][0]=n/2;

  for(int i=2;i<=11;++i)
    {
      count=0;

      for(int j=0;j<(int)pow(2,i/2-1);++j)
	{
	  if(i%2==0)
	    ret_value[i][j]=(int)((1+4*j)/pow(2,i/2+1)*n);
	  else
	    ret_value[i][j]=(int)((3+4*j)/pow(2,i/2+1)*n);
	}
    }

  return ret_value;
}

double Get_SAD(Pict *left,Pict *right,
	       int x,int y,int d,int blocksize)
{
  double tmp_error=0.0;
  int tmp_count=0;

  for(int i=-blocksize;i<=blocksize;++i)
    for(int j=-blocksize;j<=blocksize;++j)  
      {
	if(y+i<0 || y+i>left->height-1 ||
	   x+j-d<0 || x+j-d>left->width-1 ||
	   (x==0 && y==0))
	  continue;

	tmp_error+=
	  sqrt(pow((double)left->image[((y+i)*left->width*3)+((x+j)*3)+0]-
		   (double)right->image[((y+i)*left->width*3)+((x+j-d)*3)+0],2.0)+
	       pow((double)left->image[((y+i)*left->width*3)+((x+j)*3)+1]-
		   (double)right->image[((y+i)*left->width*3)+((x+j-d)*3)+1],2.0)+
	       pow((double)left->image[((y+i)*left->width*3)+((x+j)*3)+2]-
		   (double)right->image[((y+i)*left->width*3)+((x+j-d)*3)+2],2.0));

	tmp_count++;
      }

  if(tmp_count>0)
    return tmp_error/=(double)tmp_count;
  else
    return LARGE_VALUE;
}

void Block_Matching(Pict* left,Pict *right,Pict *out,
		    int max_disparity,int blocksize)
{
  int tmp_disparity=0;
  int tmp_count;
  double min_error;
  double tmp_error;

  blocksize/=2;

  for(int i=0;i<left->height;++i)
    for(int j=0;j<left->width;++j)  
      {
	min_error=16777216.0;

	for(int l=0;l<max_disparity;++l)  
	  {
	    tmp_error=Get_SAD(left,right,
				     j,i,l,blocksize);

	    if(tmp_error<min_error)
	      {
		min_error=tmp_error;
		tmp_disparity=l;
	      }
	  }
	
	out->image[i*left->width+j]=(uchar)(tmp_disparity*(255.0/(double)max_disparity));
      }
}

void Hierarchical( Pict *left,Pict *right,char* outname,
			 int max_disparity,int blocksize,
			 int **disparity_array)
{
  int tmp_disparity=0;
  int tmp_count;
  double min_error;
  double tmp_error;

  GraphType *graph;

  uchar **default_cost;  
  Pict *tmp_cost=(Pict*)malloc(sizeof(Pict));
  double **data_a;
  double **data_b;
  double smooth_aa,smooth_bb,smooth_ab,smooth_ba;

  int **tmp_g;

  double min_flow;
  double calc_flow;
  uchar success_flag;

  blocksize/=2;

  min_flow=HUGE_VALUE;

  //****ǡȤ****//
  default_cost=Alloc_2Dub(left->width,left->height);
  tmp_cost->image=Alloc_1Dub(left->width*left->height);
  data_a=Alloc_2Dd(left->width,left->height);
  data_b=Alloc_2Dd(left->width,left->height);
  tmp_g=Alloc_2Di(left->width,left->height);

#if 1
  Block_Matching(left,right,tmp_cost,
	max_disparity,blocksize*2);  

  for(int i=0;i<left->height;++i)
    for(int j=0;j<left->width;++j)    
      default_cost[i][j]=tmp_cost->image[i*left->width+j]/(192/max_disparity);
#endif

  //****Ȥι̵ʤޤǷ֤****//
  for(int h=0;h<HUGE_VALUE;++h)
    {
      //****ȹե饰****//
      success_flag=1;

      //****ƻ뺹ˤĤƷ֤****//
      for(int g=0;g<12;++g)
	{
	  //****դ****//
	  graph=new GraphType(left->width*left->height+//****ԥΡ****//
			      (left->width-1)*left->height,//****xåΡ****//
			      left->width*(left->height-1)+//****yåΡ****//
			      ((left->width-1)*left->height)*2+//****xå****//
			      (left->width*(left->height-1))*2);//****yå****//

	  //****Ρɤɲ****//
	  graph->add_node(left->width*left->height+//****ԥΡ****//
			  (left->width-1)*left->height+//****xåΡ****//
			  left->width*(left->height-1));//****yåΡ****//

	  //****ΡɤΥǡȤ****//
	  for(int i=0;i<left->height;++i)
	    for(int j=0;j<left->width;++j)  
	      {
		tmp_g[i][j]=HUGE_VALUE;
		for(int f=0;disparity_array[g][f]!=-1;++f)
		  if(abs(disparity_array[g][f]-default_cost[i][j])<
		     abs(tmp_g[i][j]-default_cost[i][j]))
		    tmp_g[i][j]=disparity_array[g][f];

		data_a[i][j]=Get_Data_Cost(Get_SAD(left,right,
								 j,i,tmp_g[i][j],blocksize));
		data_b[i][j]=Get_Data_Cost(Get_SAD(left,right,
								 j,i,default_cost[i][j],blocksize));

		graph->add_tweights(Get_Alpha_Node(0,j,i,left->width,left->height),
				    data_b[i][j],data_a[i][j]);
	      }

	  //****xå˴ؤ륳Ȥ****//
	  for(int i=0;i<left->height;++i)
	    for(int j=0;j<left->width-1;++j)
	      {
	
		smooth_aa=Get_Smooth_Cost(tmp_g[i][j],tmp_g[i][j+1]);
		smooth_bb=Get_Smooth_Cost(default_cost[i][j],default_cost[i][j+1]);
		smooth_ab=Get_Smooth_Cost(tmp_g[i][j],default_cost[i][j+1]);
		smooth_ba=Get_Smooth_Cost(default_cost[i][j],tmp_g[i][j+1]);

		//****xåΡɤΥǡȤ****//
		graph->add_tweights(Get_Alpha_Node(1,j,i,left->width,left->height),
				    smooth_bb,smooth_aa);

		if(smooth_bb<=smooth_aa)
		  {
		    graph->add_edge(Get_Alpha_Node(0,j,i,left->width,left->height),
				    Get_Alpha_Node(1,j,i,left->width,left->height),
				    Max(0,smooth_ab-smooth_bb),
				    LARGE_VALUE);
		    graph->add_edge(Get_Alpha_Node(1,j,i,left->width,left->height),
				    Get_Alpha_Node(0,j+1,i,left->width,left->height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_bb));
		  }
		else if(smooth_bb>smooth_aa)
		  {
		    graph->add_edge(Get_Alpha_Node(0,j,i,left->width,left->height),
				    Get_Alpha_Node(1,j,i,left->width,left->height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_aa));
		    graph->add_edge(Get_Alpha_Node(1,j,i,left->width,left->height),
				    Get_Alpha_Node(0,j+1,i,left->width,left->height),
				    Max(0,smooth_ab-smooth_aa),
				    LARGE_VALUE);
		  }
	      }
	  
	  //****yå˴ؤ륳Ȥ****//
	  for(int i=0;i<left->height-1;++i)
	    for(int j=0;j<left->width;++j)
	      {

		smooth_aa=Get_Smooth_Cost(tmp_g[i][j],tmp_g[i+1][j]);
		smooth_bb=Get_Smooth_Cost(default_cost[i][j],default_cost[i+1][j]);
		smooth_ab=Get_Smooth_Cost(tmp_g[i][j],default_cost[i+1][j]);
		smooth_ba=Get_Smooth_Cost(default_cost[i][j],tmp_g[i+1][j]);

		//****yåΡɤΥǡȤ****//
		graph->add_tweights(Get_Alpha_Node(2,j,i,left->width,left->height),
				    smooth_bb,smooth_aa);

		if(smooth_bb<=smooth_aa)
		  {
		    graph->add_edge(Get_Alpha_Node(0,j,i,left->width,left->height),
				    Get_Alpha_Node(2,j,i,left->width,left->height),
				    Max(0,smooth_ab-smooth_bb),
				    LARGE_VALUE);
		    graph->add_edge(Get_Alpha_Node(2,j,i,left->width,left->height),
				    Get_Alpha_Node(0,j,i+1,left->width,left->height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_bb));
		  }
		else if(smooth_bb>smooth_aa)
		  {
		    graph->add_edge(Get_Alpha_Node(0,j,i,left->width,left->height),
				    Get_Alpha_Node(2,j,i,left->width,left->height),
				    LARGE_VALUE,
				    Max(0,smooth_ba-smooth_aa));
		    graph->add_edge(Get_Alpha_Node(2,j,i,left->width,left->height),
				    Get_Alpha_Node(0,j,i+1,left->width,left->height),
				    Max(0,smooth_ab-smooth_aa),
				    LARGE_VALUE);
		  }
	      }

	  //****ήǾåȤη׻****//
	  calc_flow=graph->maxflow();

	  moge(calc_flow);

	  if(calc_flow<min_flow)
	    {
	      min_flow=calc_flow;
	      success_flag=0;

	      //****٥򹹿****//
	      for(int i=0;i<left->height;++i)
		for(int j=0;j<left->width;++j)
		  {
		    if(graph->what_segment(Get_Alpha_Node(0,j,i,left->width,left->height)) == GraphType::SOURCE)
		      default_cost[i][j]=tmp_g[i][j];
		  }
	    }

#if 0
	  for(int i=0;i<left->height;++i)
	    for(int j=0;j<left->width;++j)
	      moge(default_cost[i][j]);
#endif

	  //****դξõ****//
	  delete(graph);

	  moge(g,h);

	}

      if(success_flag==1)
	break;
    }

 Pict* out=(Pict*)malloc(sizeof(Pict));
  out->image=(uchar*)malloc(sizeof(uchar)*left->height*left->width);
  out->height=left->height;
  out->width=left->width;
  out->type=5;
  out->depth=255;

  for(int i=0;i<left->height;++i)
    for(int j=0;j<left->width;++j)
      {

	out->image[i*left->width+j]=(uchar)(default_cost[i][j]*(192.0/(double)max_disparity));  
	//	printf("%d\n",out->image[i*left->width+j]);
      }
    WriteImage_PNM(out,outname);
}


//1:Left Image 2:Right Image 3:Out Image 4:Max Disparity 5:Block Size
int main(int argc,char *argv[])
{

  Pict *left_data,*right_data,*out_data;
  int img_width;
  int img_height;

  int max_disparity;
  int block_size;

  int **disparity_array;

  char filename[256];
  //********//

    left_data=ReadImage_PNM(argv[1]);
    right_data=ReadImage_PNM(argv[2]);
    max_disparity=atoi(argv[4]);
    block_size=atoi(argv[5]);


    //****إեåѤλ뺹ΰݤ****//
    disparity_array=Calc_Disparity_Array(max_disparity);
	  moge("begin",Interval());
	  //****إեåȤԤ****//
	  Hierarchical(left_data,right_data,argv[3],
		       //out_data,
			      //img_width,img_height
			      max_disparity,block_size,disparity_array);
	  moge("end",Interval());
	  //	  WriteImage_PNM(out_data,argv[3]);

}

